use FindBin qw($Bin);
use lib $Bin;
use Getopt::Long;
use Hash::Case::Preserve;
use strict;
use warnings;
no warnings qw(uninitialized);
our $VERSION = '0.9';
print "PDKcompile.pl $VERSION by Jenda\@Krynicky.cz (c) 2004\n";

use FileHandle;
use Config::IniHash qw(ReadINI);

use Win32::Service qw(StopService StartService);

my ($help, $template, $print, %options);
GetOptions(
	'help' => \$help, 'template' => \$template,
	'h' => \$help, 't' => \$template,
	'print' => \$print, 'p' => \$print,
	'add=s' => \$options{'add'},
	'bind=s' => \$options{'bind'},
	'blib=s' => \$options{'blib'},
	'clean!' => \$options{'clean'},
	'debug=s' => \$options{'debug'},
	'dependent!' => \$options{'dependent'},
	'exe=s' => \$options{'exe'},
	'force!' => \$options{'force'},
	'gui!' => \$options{'gui'},
	'icon=s' => \$options{'icon'},
	'lib=s' => \$options{'lib'},
	'nocompress!' => \$options{'nocompress'},
	'tmpdir=s' => \$options{'tmpdir'},
	'trim=s' => \$options{'trim'},
	'verbose!' => \$options{'verbose'},
	'xclude!' => \$options{'xclude'},
);

our ($scriptfile, $config);

$scriptfile = shift();
if ($template) {
	PrintTemplate();
	exit;
} elsif ($help) {
	PrintHelp();
	exit;
} elsif (! -e $scriptfile) {
	print STDERR "Cannot find script file $scriptfile!\n\n";
	PrintHelp();
	exit;
}

my $INI = SnipINISection($scriptfile);
if (! $INI) {
	print STDERR "Cannot find\n\t=begin PDKcompile\nin script $scriptfile\n\n";
	PrintHelp();
	exit;
}

my ($ver, $filever) = findVersion($scriptfile);
$config = prepareConfig( $scriptfile, $ver, $filever);

if (! ($config = ReadINI(\$INI, heredoc => 1, withdefaults => 0, systemvars => 1,
		forValue => \&insertVars, hash => $config
	))) {
	$INI =~ s/^/\t/m;
	print STDERR "Invalid data in PDKcompile section:\n$INI\n\nGood data look like this:\n";
	PrintTemplate();
	exit;
}
clearVars();
MergeOptions($config, \%options);

my $info = FillInInfo($config->{info});
delete $ENV{Perl5opt};

print "\tcompiling $config->{main}->{name} $ver\n";

my %executable = (
	EXE => 'PerlApp.exe',
	SVC => 'PerlSvc.exe',
	DLL => 'PerlCtrl.exe',
);

my @command = (
	$executable{$config->{main}->{type}},
	OptS('add'),
	OptS('bind'),
	OptS('blib'),
	OptB('clean'),
	OptS('debug'),
	OptB('dependent'),
	OptS('exe'),
	OptB('force'),
	OptB('gui'),
	OptS('icon'),
	OptS('lib'),
	OptB('nocompress'),
	OptS('tmpdir'),
	OptS('trim'),
	OptB('verbose'),
	OptB('xclude'),
	"--info", $info,
	$scriptfile
);

if ($print) {
	print join ("\n\t", @command), "\n";

} else {

	Run($config->{'do_before'});

	print "\n";
	print join ("\n\t", @command), "\n" if $options{'verbose'};

	if (system(@command) > 0) {
		exit;
	}

	if ($config->{main}->{destination}) {
		my $source = $config->{options}->{exe};
		my $dest = (-d $config->{main}->{destination})
			? $config->{main}->{destination}.'/'.$source
			: $config->{main}->{destination};
		print "\nMoving $source to $dest\n";
		(!-e $dest or unlink $dest)
		and rename $source, $dest
		or print STDERR "Cannot move the created $source to $config->{main}->{destination}\n";
	}

	MakeHTML($config) if $config->{main}->{html};
	Run($config->{'do_after'});

}

exit();

#==========================================================
# functions

sub OptB {
	my $name = shift;
	return ($options{$name} ? "--$name" : ());
}

sub OptS {
	my $name = shift;
	return ($options{$name} ? ("--$name", $options{$name}) : ());
}

sub prepareConfig {
	my ($scriptfile, $ver, $filever) = @_;
	my ($volume, $scriptdir);
	($volume, $scriptdir, $scriptfile) = File::Spec->splitpath(File::Spec->rel2abs($scriptfile));
	$scriptdir = $volume.$scriptdir;


	my $config = {};
	tie %$config, 'Hash::Case::Preserve';
	$config->{main} = {};
	tie %{$config->{main}}, 'Hash::Case::Preserve';
	$config->{options} = {};
	tie %{$config->{options}}, 'Hash::Case::Preserve';
	$config->{info} = {};
	tie %{$config->{info}}, 'Hash::Case::Preserve';

	my ($main, $options, $info) = ( $config->{main}, $config->{options}, $config->{info});

	$main->{script} = $main->{scriptfile} = $scriptfile;
	$main->{scriptdir} = $scriptdir;
	$main->{ver} = $main->{version} = $ver;
	$main->{filever} = $main->{fileversion} = $filever;

	{
		my $name = $scriptfile;
		$name =~ s/\.([^.]+)$//; # strip extension
		my $ext = $1;
		$name = ucfirst($name);
		$main->{name} = $name;
		$config->{options}->{exe_def} = 1;
		if ($ext =~ /^(pm|ctrl)$/i) {
			$config->{options}->{exe} = "$name.DLL";
			$main->{type} = 'DLL';
		} else {
			$config->{options}->{exe} = "$name.exe";
			$main->{type} = 'EXE';
		}
	}

	$main->{pod} = $scriptfile;
	$options->{force} = 1;
	$info->{FileOS} = '0x4004';
#	$info->{FileType} = '0x1';
#	$info->{Language} = 'English';

	return $config;
}

sub findVersion {
	my $scriptfile = shift;
	my ($FILE,$ver);
	open $FILE, '< ' .  $scriptfile or die "Cannot open script file $scriptfile : $!\n";
	while (<$FILE>) {
		if (/^\s*(?:our|my)?\s*\$(?:\w+::)*VERSION\s*=\s*['"]?([0-9.]+)/i
		or /^\s*\*(?:\w+::)*VERSION\s*=\s*\\['"]?([0-9.]+)/i
		or /^\s*VERSION\s*=>\s*['"]?([0-9.]+)/i) {
			$ver = $1;
			last;
		}
	}
	close $FILE;
	die "Cannot find \$VERSION=... in the script file $scriptfile!\n"
		unless defined $ver;

	my $filever = $ver;
	$filever =~ tr/0-9.//cd; # strip anything except numbers and dots

	$filever =~ s/^\./0./;
	$filever =~ s/\.$//;
	$filever =~ s/\.\./.0./;

	if ($filever !~ s/^(\d+\.\d+\.\d+\.\d+)/$1/) {
		$filever .= '.0.0.0';
		$filever =~ s/^(\d+\.\d+\.\d+).*/$1/;
		$filever .= '.' . BuildNumber($scriptfile, $ver);
	}

	return ($ver, $filever);
}

sub SnipINISection {
	my $scriptfile = shift;
	my ( $FILE, $INI, $PDK_INI);
	open $FILE, '< ' .  $scriptfile or die "Cannot open script file $scriptfile : $!\n";
	while (<$FILE>) {
		if (/^=(?:begin|for)\s+compile\s*$/i) {
			while (<$FILE>) {
				last if /^=(?:end|cut)\s*$/;
				$INI .= $_;
			}
		}
		if (/^=(?:begin|for)\s+PDKcompile\s*$/i) {
			while (<$FILE>) {
				last if /^=(?:end|cut)\s*$/;
				$PDK_INI .= $_;
			}
		}
	}
	close $FILE;
	return $INI . $PDK_INI;
}

sub BuildNumber {
	my ($script, $ver) = @_;
	my ($line, $VER, $build);
	if (open $VER, "< $script.ver") {
		$line = <$VER>;
		close $VER;
		if ($line =~ /^\Q$ver\E : (\d+)$/) {
			$build = $1 + 1;
		} else {
			$build = 0;
		}
	} else {
		$build = 0;
	}
	if (! $print) {
		open $VER, "> $script.ver";
		print $VER "$ver : $build\n";
		close $VER;
	}

	return $build;
}

sub FillInInfo {
	our $info = shift();

	$info->{ProductName} = "$config->{main}->{name} $ver" unless $info->{ProductName};
	if (!$info->{LegalCopyright}) {
		if ($info->{CompanyName}) {
			$info->{LegalCopyright} = "$info->{CompanyName}  ".((localtime())[5] + 1900);
		}
	} else {
		$info->{LegalCopyright} =~ s/\(c\)//g;
	}
	$info->{OriginalFilename} = $options{exe} unless $info->{OriginalFilename};
	$info->{InternalName} = "$config->{main}->{name} $ver" unless $info->{InternalName};
	$info->{FileVersion} = $filever unless $info->{FileVersion};
	$info->{ProductVersion} = $filever unless $info->{ProductVersion};

	return join( ';', map { "$_=$info->{$_}" } grep {defined $info->{$_} and $info->{$_} ne ''} keys %$info);
}

sub MergeOptions {
	my ($config, $cmdline_options) = @_;

	# merge the options specified in the POD section with the ones from the command line
	foreach my $option (keys %$cmdline_options) {
		if (defined $cmdline_options->{$option}) {
			$config->{options}->{$option} = $cmdline_options->{$option};
		} else {
			$cmdline_options->{$option} = $config->{options}->{$option};
		}
	}
}

sub PrintHelp {
	print <<'*END*';
Ussage: PDKcompile [options] script.pl
    or  PDKcompile [options] control.ctrl
    or  PDKcompile --template

The script.pl or control.ctrl MUST contain a POD section
    =for PDKcompile
or
    =begin PDKcompile
The section contains data about the executable, service or COM object
to be created in INI file like format.
Run
    PDKcompile -t
to get an empty template.

The only options processed by PDKcompile itself are
    --help  : print this information
    --print : do not execute PerlApp/PerlSvc/PerlCtrl, just print the command
    --template : print a template of the POD section
all others are passed to PerlApp/PerlSvc/PerlCtrl.
*END*
}

sub PrintTemplate {
	print '=begin PDKcompile', <<'*END*';


;	You may use $variables in the values.
;	Predefined variables:
;		scriptfile = name of the file we compile (just the filename!)
;		scriptdir = full path to the directory where is the compiled script stored
;		name = the name part of the $scriptfile
;	All values specified in ANY section may be used in later values.
;	The variable names are case insensitive, as are all options, variable names
;	may contain only word characters =~ /^\$\w+$/.
;	You only get the default values if you do NOT specify the option at all!

[main]
;name=
;	Name of the project
;	By default the name of the script without extension
;type=
;	EXE, SVC or DLL
;	By default DLL for *.pm or *.ctrl, EXE otherwise
;pod=
;	What file to process with pod2html
;	By default the script
html=
;	Where to write the the HTML docs.
;	If this option is empty, no docs are created.
polishhtml=0
;	Moves the index below the name and version. Removes the link to index from AUTHOR and DISCLAIMER.
destination=
;	Where to move the created file to.

[options]
add=
;	Additional modules to include.
bind=
;	Additional files to be bound into the executable
blib=
;	Additional places to look for perl modules
clean=
;	Don't leave behind object files
debug=
;	Create a (remote) debugging executable
dependent=
;	Exe will be dependent of a Perl installation
;exe=
;	Name of exe file to generate
;	By default "$name.exe" for EXE and SVC types and "$name.dll" for DLL type.
;	You should only uncomment this if you want to create a different name!
gui=
;	Don't use a console
icon=
;	set icons associated with executable
lib=
;	additional places to look for perl modules
nocompress=
;	module source not compressed in exe
tmpdir=
;	provide a replacement dir for /tmp
trim=
;	don't include certain modules
verbose=
;	more noise when running
xclude=
;	don't include perl56.dll

[info]
CompanyName=
FileDescription=
ProductName=$name $ver
LegalCopyright=$CompanyName (c) 2002
;	in LegalCopyright option (c) is converted to the copyright sign
LegalTrademarks=
OriginalFilename=$exe
InternalName=$name $ver
FileOS=0x4004
Comments=

[do_before]
start_services=
;	Comma separated list of service names
stop_services=
run=
;	Newline separated list of commands to run. Use the HEREDOC syntax.
;	If you do not want to wait for the program to finish use this:
;		start Notepad.exe $script

[do_after]
start_services=
stop_services=
run=

=cut
*END*
}

sub Run {
	my $tasks = shift() or return;
	for ($tasks->{'stop_services'}, $tasks->{'start_services'}, $tasks->{'run'}) {
		s/^\s+//;
		s/\s+$//;
	}
	if ($tasks->{'stop_services'}) {
		print "\nStopping services:\n";
		my @services = split /\s*[;,]\s*/, $tasks->{'stop_services'};
		foreach (@services) {
			next if $_ eq '';
			print "\t$_\n";
			StopService(undef, $_);
		}
		sleep(1);
		foreach (@services) {
			next if $_ eq '';
			my %status;
			Win32::Service::GetStatus(undef, $_, \%status);
			if (! %status) {
				print "\t\tTHE SERVICE DOESN'T EXIST!\n";
				next;
			};
			while ($status{CurrentState} != 1) {
				sleep(1);
				print '.';
				Win32::Service::GetStatus(undef, $_, \%status);
			}
		}
		print " done.\n";
	}
	if ($tasks->{'start_services'}) {
		print "\nStarting services:\n";
		my @services = split /\s*[;,]\s*/, $tasks->{'start_services'};
		foreach (@services) {
			next if $_ eq '';
			print "\t$_\n";
			StartService(undef, $_);
		}
		sleep(1);
		foreach (@services) {
			next if $_ eq '';
			my %status;
			Win32::Service::GetStatus(undef, $_, \%status);
			if (! %status) {
				print "\t\tTHE SERVICE DOESN'T EXIST!\n";
				next;
			};
			while ($status{CurrentState} != 4) {
				sleep(1);
				print '.';
				Win32::Service::GetStatus(undef, $_, \%status);
			}
		}
		print " done.\n";
	}
	if ($tasks->{'run'}) {
		chomp $tasks->{'run'};
		print "\nRunning commands:\n";
		foreach (split /\n/, $tasks->{'run'}) {
			next if $_ eq '';
			print "\t$_\n";
			system($_);
		}
		print " done.\n";
	}

}

sub MakeHTML {
	my $config = shift();
	my ($name, $ver, $source, $htmlfile) = map {$config->{main}->{$_}} qw(name ver pod html);
	my $desc = $config->{info}->{FileDescription};

	print "\nCreating HTML documentation: $source => $htmlfile\n";
	system qq{pod2html --title "$name $ver : $desc" --backlink "_index_" $source > $htmlfile};

	if ($config->{main}->{polishhtml}) {
		open IN, "<$htmlfile" or die "$!\n";
		open OUT, ">$htmlfile.tmp" or die "$!\n";

		while (<IN>) {
			last if /^\Q<!-- INDEX BEGIN -->\E$/;
			print OUT $_;
		}

		my $index = $_;
		while (<IN>) {
			next if /^\t\Q<LI><A HREF="#$name">\E/i;
			$index .= $_;
			last if /^\Q<HR>\E$/;
		}

		while (<IN>) {
			last if /^\Q<A HREF="#__index__">\E/;
			print OUT $_;
		}

		<IN>;

		print OUT $index;

		while (<IN>) {
			print OUT $_;
			last if m{^\Q<H1><A NAME="author">AUTHOR</A></H1>\E$};
		}

		while (<IN>) {
			next if m{\Q<A HREF="#__index__">\E};
			print OUT $_;
		}

		close IN;
		close OUT;

		unlink $htmlfile;
		rename $htmlfile.'.tmp', $htmlfile;
	}
}

{
	my %data;
	sub insertVars {
		my ($name, $value, $section, $hash) = @_;
		$name = lc $name;
		$section = lc $section;

		$value =~ s/(?:\x0D\x0A?|\x0A)/\x0D\x0A/sg; #convert end of lines to CRLF
		$value =~ s/\$(\w+)/$data{lc $1} || $config->{main}->{$1} || $config->{options}->{$1} || $config->{info}->{$1} || '$'.$1/ge; # interpolate variables

		if ($section eq 'main' and $name eq 'exe') {
			$config->{options}->{exe_def} = 0;

		} elsif ($section eq 'main' and $name eq 'type') {
			$value = uc($value) || 'EXE';
			die "The [main]type may only be set to EXE, DLL or SVC!\n"
				if $value !~ /^(?:EXE|SVC|DLL)$/;

			if ($config->{options}->{exe_def}) {
				if ($value eq 'DLL') {
					$config->{options}->{exe} = "$config->{main}->{name}.dll";
				} else {
					$config->{options}->{exe} = "$config->{main}->{name}.exe";
				}
			}

		} elsif ($section eq 'main' and $name eq 'name') {
			if ($config->{options}->{exe_def}) {
				if ($config->{main}->{type} eq 'DLL') {
					$config->{options}->{exe} = "$value.dll";
				} else {
					$config->{options}->{exe} = "$value.exe";
				}
			}

		} elsif ($section eq 'main' and $name eq 'html' and $value eq '1') {
			$value = "$config->{main}->{name}.html";
		}

		$data{$name} = $value;
		return $value;
	}

	sub clearVars {undef %data}
}

1;

__END__
=begin compile

[main]
name=PDKCompile
;	Name of the project
type=EXE
;	EXE, SVC or DLL
;pod=
;	What file to process with pod2html
html=
;	Where to write the the HTML docs.
;	If this option is empty, no docs are created.
polishhtml=0
;	Moves the index below the name and version. Removes the link to index from AUTHOR and DISCLAIMER.
destination=d:\perl
;	Where to move the created file to.

[options]
add=
;	Additional modules to include.
bind=
;	Additional files to be bound into the executable
blib=
;	Additional places to look for perl modules
clean=
;	Don't leave behind object files
debug=
;	Create a (remote) debugging executable
dependent=
;	Exe will be dependent of a Perl installation
;exe=
;	Name of exe file to generate; by default "$name.exe" for EXE and SVC types and "$name.dll" for DLL type.
gui=
;	Don't use a console
icon=$scriptdir$name-nix.ico
;	set icons associated with executable
lib=
;	additional places to look for perl modules
nocompress=
;	module source not compressed in exe
tmpdir=
;	provide a replacement dir for /tmp
trim=
;	don't include certain modules
verbose=
;	more noise when running
xclude=
;	don't include perl56.dll

=cut

=begin PDKcompile

[options]
icon=$scriptdir$name.ico

[info]
CompanyName=Jenda Krynicky -- http://Jenda.Krynicky.cz/
FileDescription=PDKCompile - Perl Development Kit (PDK) Compiler Frontend
ProductName=$name $ver
LegalCopyright=Copyright (c) 2002 Jenda Krynicky <Jenda@Krynicky.cz>. All rights reserved.
LegalTrademarks=n.a.
OriginalFilename=$exe
InternalName=$name $ver
FileOS=0x4004
Comments=<<*END*
PDKCompile, the Perl Development Kit (PDK) Compiler Frontend, is a wrapper application for ActiveState's (http://www.activestate.com/) Perl compilers.
PDKCompile manages all PerlApp/PerlSvc/PerlCtrl compile tasks using instructions embedded in source files in a POD section.
This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.
*END*

[do_before]
start_services=
;	Comma separated list of service names
stop_services=
run=
;	Newline separated list of commands to run. Use the HEREDOC syntax.
;	If you do not want to wait for the program to finish use this:
;		start Notepad.exe $script

[do_after]
start_services=
stop_services=
run=

=cut

